/*
 * IRC - Internet Relay Chat, ircd/p_sethost.c
 * Copyright (C) 1990 Jarkko Oikarinen and
 *                    University of Oulu, Computing Center
 *
 * See file AUTHORS in IRC package for additional names of
 * the programmers.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "channel.h"
#include "config.h"
#include "client.h"
#include "ircd.h"
#include "ircd_defs.h"
#include "ircd_string.h"
#include "ircd_alloc.h"
#include "list.h"
#include "msg.h"
#include "numeric.h"
#include "numnicks.h"
#include "patch.h"
#include "querycmds.h"
#include "send.h"
#include "s_conf.h"
#include "s_user.h"
#include "userload.h"

#include <string.h>


/*=============================================================================
 * Proto types
 */
static char *FindNVhost(int hostnum);
static unsigned int FindSVhost(char *hostmsk);


/* --------------------------------------------------------
 * **Daz** m_sethost()
 *
 * SETHOST allows users/opers (provided the necessary
 * feature(s) are enabled - to change their host to that
 * specified in an S:line in the ircd.conf file.
 *
 * It is the system admins responsibility to ensure that
 * S:line hosts do not break any trademark or copyright laws,
 * since the ircd will accept them "as is".
 *
 * In order for users (ie. non-opers) to use SETHOST, the
 * 'USER_SETHOST' feature must be enabled.  Additionally there
 * must be one or more S:line hosts with password fields
 * specified (a password is essential for user sethosts).
 *
 * Opers can (assuming OPER_DYNHOST is enabled) set any
 * S:line they choose (including ones designated for users)
 *
 * Syntax (for users):
 *      /quote sethost <hostmask> <password>
 *   or /quote sethost <#number> <password>
 *           ..where <number> is the number shown on /stats s
 *
 * Note: users *cannot* change their ident, nor specify a
 *       quit reason.
 *
 * Syntax: (for opers):
 *      /quote sethost <ident> <hostmask> <reason for quit>
 *   or /quote sethost <ident> <#number> <reason for quit>
 *
 *
 * --------------------------------------------------------
 */

/* m_sethost
 * 
 *      parv[0] = sender prefix
 *      parv[1] = new username
 *      parv[2] = new host/number
 *      parv[parc - 1] = quit reason (optional)
 */
int m_sethost(struct Client *cptr, struct Client *sptr, int parc, char *parv[])
{
  struct Membership *lp;
  struct SLink *lp2;
  struct Channel *chptr;
  struct User *user = sptr->user;
  unsigned int target;
  char tempip[17];
  char ip_base64[8];
  char *str, *nick, *newhost, *newuser, *comment;
  char *chandb;
  int restore = 0;  
              
  /* Is option switched off currently? */
  if (!IsOption(OPER_DYNHOST) && !IsOption(USER_SETHOST))
    return 0;
  else if (!IsAnOper(sptr) && !IsOption(USER_SETHOST)) {
    /* SETHOST enabled for opers only, and user isn't an oper */
    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
    return 0;
  }
              
  /* We need at least two arguments whatever the command is */
  if (parc < 2) {
     sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SETHOST");
     return 0;
  }

  nick = parv[0];
  if (!strCasediff(parv[1],"undo")) {
     /* User wishing to undo sethost, make sure we can.. */
     if (IsSetHost(sptr)) {
        newuser = (sptr->user->origuser[0]!='\0')?sptr->user->origuser:sptr->user->username;
        newhost = sptr->user->orighost;
        restore = 1;
     } else
        return 0;       /* Cannot undo */
  } else {
     /* Make sure we have the necessary arguments */
     if (parc < 3) {
        sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SETHOST");
        return 0;
     }
     if (IsAnOper(sptr)) {
        newuser = parv[1];
        newhost = parv[2];
     } else {
        newuser = sptr->user->username; /* Users not allowed to change IDENT */
        newhost = parv[1];
     }
     restore = 0;
  }
 
  /* Note: we only perform the SETHOST checks when the user wishes to change *TO*
   * a host other than the one they connected to IRC on (ie. their original host).
   * Since we know for them to connect in the first place, their username and hostmask
   * must've been syntactically correct - so there is no point at all in checking
   * it here.. so we just skip it.
   */
  if (restore == 0) {
     /* Check lengths don't exceed respective MAXlengths */
     if (strlen(newuser) > USERLEN) {
        sendto_one(sptr, err_str(ERR_UHOSTBADSIZE), me.name, parv[0], "SETHOST",
           "Username", "long", USERLEN);
        return 0;
     }
     else if (strlen(newuser) <= 0) {
        sendto_one(sptr, err_str(ERR_UHOSTBADSIZE), me.name, parv[0], "SETHOST",
           "Username", "short", USERLEN);
       return 0;
     }
  
     /* We branch depending on whether the user types:
      * /quote sethost <user> <host> ... -or-
      * /quote sethost <user> <#number> ...
      */
     if (newhost[0] == '#') {
        target = atoi(newhost + 1);
              
        newhost = FindNVhost(target);
        if (!newhost) {
           sendto_one(sptr, err_str(ERR_VHOSTUNAVAIL), me.name, parv[0], "SETHOST");
           return 0;
        }
     }
     else {  
        /* Couldn't find textual vhost */
        if (!FindSVhost(newhost)) {
           sendto_one(sptr, err_str(ERR_VHOSTUNAVAIL), me.name, parv[0], "SETHOST");
           return 0;
        }
     }
      
     if (strlen(newhost) > HOSTLEN) {
        sendto_one(sptr, err_str(ERR_UHOSTBADSIZE), me.name, parv[0], "SETHOST",
           "Hostmask", "long", HOSTLEN);
        return 0;
     }
     else if (strlen(newhost) <= 0) {
        sendto_one(sptr, err_str(ERR_UHOSTBADSIZE), me.name, parv[0], "SETHOST",
           "Hostmask", "short", HOSTLEN);
        return 0;
     }
     
     /* Check the hostmask/username here for syntactic correctness */
     if (!(IsValidHost(newhost)) || !IsValidHost(newuser)) {
        sendto_one(sptr, err_str(ERR_BADHOSTMASK), me.name, parv[0], "SETHOST");
        return 0;
     }
  
     /* Check to see whether the SETHOST is Oper only (no password) or
      * a user-settable S:line (password defined) - and check to see password
      * is correct! :)
      */
     if (comment = GetVhostPass(newhost)) {
        /* Password defined - check it (only if user isn't an Oper) */
        if (strCasediff(comment,parv[2]) && !IsAnOper(sptr)) {
           sendto_one(sptr, err_str(ERR_PASSWDMISMATCH), me.name, nick);
        
           /* Broadcast failed SETHOST attempt */
           LogFeature(sptr, USER_SETHOST, "failed SETHOST attempt for host '%s' (bad password)", newhost);
           return 0;
        }
     } else {
        /* User isn't allowed to use this SETHOST */
        if (!IsAnOper(sptr)) {
           sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
           return 0;
        } else {
           /* Option isn't enabled - Oper can't set unpassworded S:lines */
           if (!IsOption(OPER_DYNHOST))
              return 0;
        }
     }
     
     /* As a final check - if an Oper is SETHOST'ing to a passworded host, make
      * sure they dont accidentally output the password as their 'quit' comment
      * (since opers dont need to use the password)..
      */
     if (comment && IsAnOper(sptr))
        comment = "host change";
  }   

  /* Propogate changes across connected servers...
           
   * In order to 'fool' the other servers (which may be unpatched)
   * into accepting the new host, we have to PART from all channels,
   * issue a QUIT to other servers (although you will still appear
   * to be connected) and send the new NICK to all servers.
   *
   * If we don't QUIT the user in the usual manner, non-patched servers
   * will Nick Collide the new instance of the client, as it will appear
   * to have a different user@host to the 'other' client (which is still
   * seen as connected).
   *          
   * We also issue a PART message for all channels the user is on.  This
   * message is propogated to the user issuing the command ONLY.  Since
   * we are sending the QUIT message out to the remote servers/clients, there
   * is no need to issue PART messages to them as well.  The reason
   * we do this is entirely to ensure client sync - there is no effect on
   * network sync as the client will be QUIT'd in the usual manner.
   *
   * Effectively the command works like this:
   *      - client issues /quote sethost command
   *      - PART messages for all channels are sent to the client ONLY, it
   *        will internally appear to be no longer on said channels, although
   *        on remote servers he/she will be still present (this doesn't matter
   *        as it's handled by the later QUIT message)
   *      - QUIT message is then propogated to connected servers, which can
   *        then update their own internal hashtables for the client (this is
   *        outside the client scope, so we aren't too concerned about it)
   *      - client will then appear as if he/she has just connected to IRC, will
   *        not be on any channels, but will have the newly spoofed user@host.
   */ 
     
  /* Remove records of the client from channels, cancels any ping
   * requests pending (they wont return).
   */
      
  comment = (parc > 3) ? parv[parc - 1] : "host change";
     
  if (IsUser(sptr)) {
    /*
     * clear out uping requests
     */
    if (IsUPing(sptr))
      uping_cancel(sptr, 0);

    /* Stop a running /LIST clean */
    if (MyUser(sptr) && sptr->listing) {
      sptr->listing->chptr->mode.mode &= ~MODE_LISTED;
      MyFree(sptr->listing);
      sptr->listing = NULL;
    }
      
    /*
     * If a person is on a channel, send a QUIT notice
     * to every client (person) on the same channel (so
     * that the client can show the "**signoff" message).
     * (Note: The notice is to the local clients *only*)
     */
    sendto_common_channels_butone(sptr, ":%s QUIT :%s", sptr->name, "host change");
    /* PART user from all channels he/she is currently on
     * (this is only sent to the client)
     */  
    chandb = (char *) MyMalloc((sptr->user->joined * (CHANNELLEN + 1)) + 1);
    if (!chandb)
       return 0;
    else 
       *chandb = '\0';
    if(lp = sptr->user->channel)
    {
      for (; lp->next_channel; lp = lp->next_channel);
      for (; lp; lp = lp->prev_channel) {
         chptr = lp->channel;
         strcat(chandb,chptr->chname);    /* Cache channel (for readding later) */
      
         strcat(chandb,",");              /* Note: ',' is an illegal char in both nicks and
                                         * channel names - so therefore it is used here
                                         * as a delimiter char to seperate channels
                                         * (ie. <chan1>,<chan2>)
                                         */
  
       /* Note: we don't send the PART message to the client if QUICK_SETHOST
        * is defined, but we *must* remove the user from the channel in this
        * loop - since otherwise it would never terminate...
        */
          sendto_one(sptr, ":%s!%s@%s PART %s", nick, sptr->user->username,
             sptr->user->host, chptr->chname);
          remove_user_from_channel(sptr, chptr);
      }
    }            
    /* Clean up invitefield */
    while ((lp2 = sptr->user->invited))
      del_invite(sptr, lp2->value.chptr);
   
    /* Clean up silencefield */
    while ((lp2 = sptr->user->silence))
      del_silence(sptr, lp2->value.cp);
   
    /* Note at this point, in order to keep network client-totals we have to add
     * some values that have at this stage been subtracted from 'UserStats'
     * If we don't do this we end up with very odd minus numbers.. :)
     */
    ++UserStats.local_clients;
    ++UserStats.clients;
   
    if (MyConnect(sptr))
      Count_clientdisconnects(sptr, UserStats);
    else
      Count_remoteclientquits(UserStats, sptr);
  }  
  
  /* WALLOP's change of host (if applic.) */
  if(IsAnOper(sptr))
  {
    if (LogOption(OPER_DYNHOST) && MyUser(sptr))
       LogFeature(sptr, OPER_DYNHOST, "changed his/her host to '%s@%s'", newuser, newhost);    
  }
  else
  {
    if (LogOption(USER_SETHOST) && MyUser(sptr))
       LogFeature(sptr, USER_SETHOST, "activated user sethost for '%s@%s'", newuser, newhost);    
  }
  /* Backup the user's host (for undoing purposes) */
  if (restore == 0 && !IsSetHost(sptr))
  {
     /* Backup user's REAL host (we might need it later for undoing purposes) */
     ircd_strncpy(sptr->user->orighost, sptr->user->host, sizeof(sptr->user->orighost) - 1);
     /* And also ident */
     ircd_strncpy(sptr->user->origuser, sptr->user->username, sizeof(sptr->user->origuser) - 1);
  }  
  /* Ok so far, lets copy the newhost/newuser variables over the 'real' ones */
  ircd_strncpy(sptr->user->host, newhost, sizeof(sptr->user->host) - 1);
  ircd_strncpy(sptr->sockhost, newhost, sizeof(sptr->sockhost) - 1);
  ircd_strncpy(sptr->user->username, newuser, sizeof(sptr->user->username) - 1);
     
  if (restore == 1)
  {
     sptr->user->orighost[0] = '\0';
     sptr->user->origuser[0] = '\0';
  }
  /* Spoof a QUIT message to be sent to all connected servers
   * This will tell all clients that you've 'QUIT', with reason:
   * 'Host change' (or whatever 'comment' may be)
   */   
  sendto_serv_butone(cptr, "%s%s " TOK_QUIT " :%s", NumNick(sptr), comment);
     
  /* Now 'CREATE' the new client with the new hostmask.  Note that
   * at this point non-patched servers will treat this like a new client
   * joining, which is exactly how they should treat it :)
   */
  str = umode_str(sptr);
      
  sendto_serv_butone(cptr, *str ?
      "%s " TOK_NICK " %s %d %d %s %s +%s %s %s%s :%s" :
      "%s " TOK_NICK " %s %d %d %s %s %s%s %s%s :%s",
      NumServ(user->server), nick, sptr->hopcount + 1, sptr->lastnick,
      user->username, user->host, str, inttobase64(ip_base64,
      ntohl(sptr->ip.s_addr), 6), NumNick(sptr), sptr->info);
   
  /* Output success message */
  if (restore == 0) {
     sendto_one(sptr, rpl_str(RPL_NEWHOSTIS), me.name, nick, "SETHOST",
        "Your new", "is", newuser, newhost);
//     SetSethost(sptr);
  } else {
     sendto_one(sptr, rpl_str(RPL_NEWHOSTIS), me.name, nick, "SETHOST",
        "Restored your", "to", user->username, sptr->user->host);
//     ClearSethost(sptr); 
  }

  /* Rejoin user to channel(s) (if applicable) */
  RejoinChannels(sptr, chandb);

  MyFree(chandb);

  return 0;
}


/* --------------------------------------------------------
 * **Daz** FindNVhost()
 *
 * Takes a number as a parameter (>= 1) and finds the S:line
 * vhost at that position.
 *
 * Returns: the hostmask or NULL (an invalid hostmask char)
 * if no matching hostmask for the given number was found
 * --------------------------------------------------------
 */
static char *FindNVhost(int hostnum)
{
  struct ConfItem *aconf;
  unsigned int i;
 
  for (i = 0, aconf = GlobalConfList; aconf; aconf = aconf->next)
    if (aconf->status == CONF_VHOST && hostnum == ++i)
      return aconf->host;
     
  return NULL;
}
       
          
/* --------------------------------------------------------
 * **Daz** FindSVhost()
 *
 * Takes a hostmask string as a parameter, searches through
 * the specified S:lines, returning '1' if the hostmask
 * was found, '0' if not..
 * --------------------------------------------------------
 */
static unsigned int FindSVhost(char *hostmsk)
{
  struct ConfItem *aconf;
 
  for (aconf = GlobalConfList; aconf; aconf = aconf->next)
    if (aconf->status == CONF_VHOST && strCasediff(hostmsk, aconf->host) == 0)
      return 1;
 
  return 0;
}
